from typing import Dict, Iterable, List, Optional, cast

from isahitlab.actions.base import BaseAction
from isahitlab.api.project_configuration.api import ProjectConfigurationApi
from isahitlab.api.task.api import TaskApi
from isahitlab.domain.batch import BatchId
from isahitlab.domain.task import (TaskCompatibilityMode, TaskFilters, TaskId,
                                   TaskOptionalFields, TaskStatus)
from isahitlab.formatters import get_compatibility_formatter
from isahitlab.operations.project_configuration.get_project_configuration import \
  GetProjectConfigurationOperation
from tqdm import tqdm
from typeguard import typechecked


class ProjectConfigurationActions(BaseAction):
    """Project configuration actions"""

    @typechecked
    def project_configuration(
        self,
        project_id: str,
        disable_progress_bar: Optional[bool] = False
    ) -> Dict:
        """ Get the configuration of a project
        
        Args:
            project_id: ID of the project
            disable_progress_bar: Disable the progress bar display

        Returns:
            Project configuration representation
        """

        operation = GetProjectConfigurationOperation(self.http_client)

        return operation.run(project_id=project_id, disable_progress_bar=disable_progress_bar)

    


import json
import logging
import pathlib
import uuid
from datetime import datetime
from typing import Any, Dict, List, Literal, Union

from isahitlab.domain.project import ProjectType
from isahitlab.helpers.labels import extract_main_labels
from isahitlab.helpers.path import check_extension
from isahitlab.helpers.points import (bbox_to_corners, denormalize_points,
                                      normalize_points, vertices_to_points)

from .base import BaseFormatter

logger = logging.getLogger(__name__)

class LabToKiliFormatter(BaseFormatter):
    """Formatter used to ensure retrocompatibilty with kili"""

    is_sequence: bool = False

    export : List[Dict[str, Any]]= None

    def __init__(self, project_configuration : Dict, options : Dict = {}) -> None :

        self._init_configuration(project_configuration, options, {
            "compatible_types" : [
                "form.tool-iat-rectangle", 
                "iat-rectangle",
                "form.tool-iat-polygon", 
                "iat-polygon",
                "form.tool-iat-polyline", 
                "iat-polyline",
                "form.tool-iat-graph", 
                "iat-graph",
                "form.tool-iat-segmentation", 
                "iat-segmentation"
            ]
        })
        
        # Check output filename
        if options.get("output_filename", None) and not check_extension(options.get("output_filename", ""), ".json"):
            raise Exception('"output_filename" must have .json extension')
        
        self.is_sequence = self._project_configuration.get("metadata", {}).get("toolOptions", {}).get("sequence", False)
        

    ### Transform

    def format_tasks(self, tasks: List[Dict]) -> List[Dict]:
        formatted_tasks = []

        for task in tasks:
            formatted_task = self._format_task(task)
            formatted_tasks.append(formatted_task)

        return formatted_tasks
    
    ### Export

    def _init_export(self):
        self.export = []
    
    def append_task_to_export(self, task: Dict):
        if self.export == None:
            self._init_export()

        self.export.append(self._format_task(task))

    def complete_export(self):
        if self.export == None:
            logger.warning('Nothing to export')
            return
        
        if self._options.get("in_memory"):
            return self.export
        
        now_str = datetime.now().strftime("%Y-%m-%d_%H:%M:%S")
        default_folder = '.'
        extra_filename_part = "_{}".format(self._options.get("extra_filename_part")) if "extra_filename_part" in self._options else ""
        default_filename = f'lab_kili{extra_filename_part}_{now_str}.json'
        output_path = pathlib.Path(self._options["output_folder"] or default_folder)
        export_name = self._options["output_filename"] or default_filename

        outputfile_path = output_path / export_name
        outputfile_path.parent.mkdir(parents=True, exist_ok=True)

        with outputfile_path.open('w') as ef:
            json.dump(self.export, ef, indent=2)

        print("Exported to \"{}\"".format(str(pathlib.Path(outputfile_path).absolute())))

    def _format_task(self, task : Dict) -> Dict:
        task = self._normalize_task(task)
        formatted_task = {
            "createdAt": task["createdAt"],
            "content": self._map_resources(task),
            "externalId": task["name"],
            "id": task["_id"],
            "isHoneypot": False,
            "jsonMetadata": {},
            "latestLabel": self._map_data(task),
            "resolution": self._map_image_size(task),
            "skipped": False,
            "status": self._map_status(task)
        }

        return formatted_task

    def _map_image_size(self, task: Dict) -> Union[Dict[str, int], None]:
        return task.get('data', {}).get('body', {}).get('image', None)

    def _map_status(self, task: Dict) -> Literal["TODO", "ONGOING", "LABELED", "REVIEWED", "TO_REVIEW"]:
        if task['status'] == 'complete':
            return 'LABELED'
        elif task['status'] == 'reviewed':
            return 'REVIEWED'
        elif task['status'] == 'to-review':
            return 'ONGOING'
        else:
            return 'TODO'


    def _map_resource(self, resource) -> Union[Dict, str]:
        if isinstance(resource, dict):
            return resource['name'] if 'name' in resource else resource['id']
        else:
            return resource

    def _map_resources(self, task: Dict) -> str:
        if "resources" in task and len(task["resources"]) > 0 :
            return ",".join([*map(lambda r : self._map_resource(r), task["resources"])])
        else:
            return ''

    def _map_data(self, task: Dict) -> Union[Dict, None]:
        if "data" not in task:
            return None
        
        data = task.get("data", None)

        if not data or len(data) == 0:
            return None

        return {
            "author": self._map_data_author(data),
            "createdAt": data['createdAt'],
            "isLatestLabelForUser": True,
            "isSentBackToQueue": False,
            "jsonResponse": self._map_data_body(task),
            "labelType": self._map_data_type(data),
            "modelName": None
        }
    
    def _map_data_author(self, data: Dict):
        """
        {
            "email": "julien.cachou.ext@veolia.com",
            "firstname": "Julien",
            "id": "clraqptja02zr087wdp9zeiu5",
            "lastname": "CACHOU",
            "name": "Julien CACHOU"
        }
        """

        if "user" not in data or not data["user"]:
            return None
        
        name = data['user']['name']

        name_parts = name.split(' ')
        firstname = name_parts[0] if len(name_parts) > 0 else name
        lastname = ' '.join(name_parts[1:]) if len(name_parts) > 1 else ''

        return {
            "email" : data['user']['email'],
            "name" : name,
            "firstname" : firstname,
            "lastname" : lastname,
            "id" : data['user']['id']
        }
    
    def _map_data_type(self, data: Dict) -> str:
        if data['type'] == 'review':
            return "REVIEW"
        else: 
            return "DEFAULT" 

    def _map_data_body(self, task: Dict) -> Union[Dict, None] :
        data = task["data"]
        if "body" not in data:
            return None
        
        body : Dict = data["body"]
        project_type : ProjectType = data["projectType"]
        resources = task.get("resources", [])

        annotations = {}

        if project_type == 'iat-rectangle':
            annotations = self._format_polygon_annotations(body, "rectangle", sequence_number=len(resources) if self.is_sequence else None)
        elif project_type == 'iat-polygon':
            annotations = self._format_polygon_annotations(body, "polygon", sequence_number=len(resources) if self.is_sequence else None)
        elif project_type == 'iat-segmentation':
            annotations = self._format_segmentation_annotations(body)
        else:
            logging.warning('LAB To kili annotations formatting is not available for project type "{}"'.format(project_type))

        return { "OBJECT_DETECTION_JOB": { "annotations" : annotations } }
        
    def _format_polygon_annotations(self, body, type: str, sequence_number: Union[int, None] = None) -> Union[List[Dict], Dict[str, Dict]]:
        formatted_annotations_by_timestamp = {}
        annotations = body["annotations"] if "annotations" in body else []

        # Inititalize dict
        if sequence_number != None:
            for s in range(0, sequence_number):
                formatted_annotations_by_timestamp[s] = {}
        
        formatted_annotations_by_timestamp["none"] = {}


        if "image" not in body or not body["image"]:
            if len(annotations) > 0:
                logging.warning("Annotations ignored due to missing image size")
            return []


        image_size = body.get("image", None)

        for annotation in annotations:
            polygons = annotation["polygons"]
            labels = annotation["labels"] if "labels" in annotation else {}
            main_labels = extract_main_labels(labels)
            formatted_annotation = {
                "boundingPoly" : [],
                "categories" : [*map(lambda main_label : { "name" : main_label['id'] }, main_labels)],
                "children" : self._format_labels(labels, True),
                "type" : type
            }
            bounding_polys_by_timestamp = {}
            for polygon in polygons:
                vertices = polygon["geometry"]["vertices"]
                if type == "rectangle":
                    vertices = bbox_to_corners(vertices)
                timestamp = int(polygon.get('timestamp', 1)) - 1  if sequence_number != None else "none"
                if timestamp not in bounding_polys_by_timestamp:
                    bounding_polys_by_timestamp[timestamp] = []
                poly = self._format_vertices(vertices, image_size)
                bounding_polys_by_timestamp[timestamp].append(poly)

            for sequence in bounding_polys_by_timestamp:
                if len(formatted_annotations_by_timestamp[sequence]) == 0:
                    formatted_annotations_by_timestamp[sequence]["OBJECT_DETECTION_JOB"] = {
                        "annotations" : []
                    }
                formatted_annotations_by_timestamp[sequence]["OBJECT_DETECTION_JOB"]["annotations"].append({ **formatted_annotation, "boundingPoly" : bounding_polys_by_timestamp[sequence] })
        
        if sequence_number != None:
            del formatted_annotations_by_timestamp["none"]
            return formatted_annotations_by_timestamp
        else:            
            return formatted_annotations_by_timestamp["none"]

    
    def _format_segmentation_annotations(self, body) -> List[Dict]:
        formatted_annotations = []
        annotations = body["annotations"] if "annotations" in body else []
        
        if "image" not in body or not body["image"]:
            if len(annotations) > 0:
                logging.warning("Annotations ignored due to missing image size")
            return []


        image_size = body.get("image", None)

        for annotation in annotations:
            polygons = annotation.get('approximation',{}).get('polygons',[])
            minus_polygons = annotation.get('approximation',{}).get("minus", [])
            labels = annotation["labels"] if "labels" in annotation else {}
            main_labels = extract_main_labels(labels)
            formatted_annotation = {
                "categories" : [*map(lambda main_label : { "name" : main_label['id'] }, main_labels)],
                "children" : self._format_labels(labels, children_only=True),
                "type" : "semantic",
                "mid" : str(uuid.uuid4()).replace('-', '')[0:15]
            }

            # Prepare minus boxes
            formatted_minus = []
            for polygon in minus_polygons:
                vertices = polygon["geometry"]["vertices"]
                formatted_minus.append(self._format_vertices(vertices, image_size, denormalized=True))

            for pindex, polygon in enumerate(polygons):
                vertices = polygon["geometry"]["vertices"]
                holes = polygon["geometry"].get("holes", [])

                poly = self._format_vertices(vertices, image_size, denormalized=True)
                if len(formatted_minus) > 0: # Old system with minus polygon not mapped to the parent polygon
                    formatted_annotations.append({
                        "boundingPoly" : [
                            poly, 
                            *(formatted_minus if pindex == 0 else [])], # We add minus polygons in the first "annotation" only
                        **formatted_annotation
                    })
                else:  # New system with holes mapped to the parent polygon
                    formatted_annotations.append({
                        "boundingPoly" : [
                            poly, 
                            *[self._format_vertices(hole, image_size, denormalized=True) for hole in holes]], # We add holes
                        **formatted_annotation
                    })
                    

        return formatted_annotations
    

    def _format_vertices(self, vertices : List[Union[float, int]], image_size : Dict[str, int], denormalized : bool = False):
        if denormalized:
            denormalized_points = vertices_to_points(vertices)
            normalized_points = normalize_points(denormalized_points, image_size)
        else:
            normalized_points = vertices_to_points(vertices)
            denormalized_points = denormalize_points(normalized_points, image_size, 2)
        return {
            "normalizedVertices" : normalized_points,
            "vertices" : denormalized_points
        }

    def _format_labels(self, labels : Dict, children_only : bool = False):
        children = {}
        for list in labels:
            if not children_only:
                children[list] = {
                    "categories" : [ { "name" : l["id"] } for l in (labels[list].get("labels", []) or [])],
                }
            if "children" in labels[list] and len(labels[list]["children"]):
                formatted_children = self._format_labels(labels[list]["children"])
                children = { **children, **formatted_children }
                
        return children
    

    def prepare_import(self, tasks: List[Dict]):
        raise NotImplementedError


import logging
from typing import List, Optional

from isahitlab.api.task.api import TaskApi
from isahitlab.domain.task import TaskId
from isahitlab.operations.base import BaseAction
from tqdm import tqdm
from typeguard import typechecked
from isahitlab.helpers.list import divide_chunks

logger = logging.getLogger('isahitlab.replay_webhooks')


class ReplayWebhooksOperation(BaseAction):
    """Replay webhooks
    """

    _task_api : TaskApi

    @typechecked
    def run(
        self,
        task_id_in: List[TaskId],
        disable_progress_bar: Optional[bool] = False,
    ) -> int:
        """ Replay webhooks
        
        Args:
            task_id_in: List of the IDs of the tasks to redo
            disable_progress_bar: Disable the progress bar display


        """
        self._task_api = TaskApi(self._http_client)


        chunks = divide_chunks(task_id_in, 100)
                
        with tqdm(total=len(task_id_in),  disable=disable_progress_bar, desc="Register webhooks for replaying... ") as loader:
            for chunk in chunks:
                self._task_api.replay_webhooks( task_ids=chunk )
                loader.update(len(chunk))
        
        return len(task_id_in)
    


""" Domain Filter to API params"""
from isahitlab.domain.integration import IntegrationFilters

def integration_filters_mapper(filters: IntegrationFilters):
    params = {
        "pagination" : "false" if filters.pagination == False else "true",
        "page" : filters.page,
        "limit" : filters.limit
    }

    if filters.visibility_in and len(filters.visibility_in):
        params['visibility'] = filters.visibility_in

    if filters.search and len(filters.search):
        params['search'] = filters.search
        
    return params


from .lab_iat_schema import LAB_IAT_SCHEMA
import copy

LAB_IAT_SEGMENTATION_SCHEMA = copy.deepcopy(LAB_IAT_SCHEMA)

# Enforce number of vertices
LAB_IAT_SEGMENTATION_SCHEMA["properties"]["annotations"]["items"]["properties"]["polygons"] = { 
    "type": "array", 
    "prefixItems": [
        { "type" : "number" },
        { "type" : "number" },
        { "type" : "number" },
    ]
}

# Add mask
LAB_IAT_SEGMENTATION_SCHEMA["properties"]["mask"] = {
    "type" : "string",
    "pattern": "^data:image/png;base64,"
}


from typing import BinaryIO, Dict, Generator
import time

from isahitlab.api.base import BaseApi
from isahitlab.domain.dataset import DatasetPayload, DatasetBrowsingFilters, FileBrowsing
from isahitlab.domain.project import ProjectId
from .mappers import dataset_browsing_filters_mapper

from ..helpers import get_response_json, log_raise_for_status


class DatasetApi(BaseApi):
    """Dataset API Calls"""

    def browse_dataset(self, filters: DatasetBrowsingFilters) -> Generator[str, None, None] :
        """Browse files"""

        while True:

            paginated: FileBrowsing = self._browse_dataset_page(filters)
            docs = paginated['files']
            filters.next_token = paginated.get('nextToken', None)
            loaded = len(docs)
            yield (docs, loaded)

            if 'nextToken' not in paginated:
                break

            time.sleep(0.5)
    
    
    def _browse_dataset_page(self, filters: DatasetBrowsingFilters) -> FileBrowsing :
        """Get file keys"""
        
        tasks = self._http_client.get('api/file-manager/datasets/{}/list'.format(filters.dataset_id), params=dataset_browsing_filters_mapper(filters))
        
        log_raise_for_status(tasks)
        
        return get_response_json(tasks)
    

    def upload_file(self, project_id: str, dataset_id: str, file : BinaryIO, folder: str) -> Dict :
        """Upload file to dataset"""
        
        files = {'file': file }

        uploaded = self._http_client.post('api/file-manager/datasets/{}/resources'.format(dataset_id), files=files, params={ "projectId" : project_id }, data={ "path": folder })
        
        log_raise_for_status(uploaded)
        
        return get_response_json(uploaded)
    
    
    def create_dataset(self, project_id : ProjectId, dataset: DatasetPayload):
        """Create a dataset in a project"""

        data = {
            "name": dataset.name,
            "projectId" : project_id
        }

        if dataset.integration_id:
            data['integration'] = dataset.integration_id
            data['key'] = dataset.base_folder


        result = self._http_client.post('api/file-manager/datasets', json=data)

        log_raise_for_status(result)

        return get_response_json(result)



import json


class AuthenticationFailed(Exception):
    def __init__(self, access_id, secret_key) -> None:
        if access_id is None or secret_key is None:
            super().__init__("You need to provide an access id and its associated secret key to connect.")
        else:
            super().__init__("Connection to isahit lab failed. Check your access id and secret key.")

class ApiBadRequest(Exception):
    def __init__(self, message) -> None:
        try:
            parsed = json.loads(message)
            message = json.dumps(parsed, indent=2)
            self.errorCode = parsed.get("errorCode", None)
        except:
            # IGNORE
            pass
        super().__init__(message)

class ApiNotFound(Exception):
    def __init__(self, message) -> None:
        try:
            message = json.dumps(json.loads(message), indent=2)
        except:
            # IGNORE
            pass
        super().__init__(message)

class ApiForbidden(Exception):
    def __init__(self, message) -> None:
        try:
            message = json.dumps(json.loads(message), indent=2)
        except:
            # IGNORE
            pass
        super().__init__(message)

class InternalServerError(Exception):
    def __init__(self, message) -> None:
        try:
            message = json.dumps(json.loads(message), indent=2)
        except:
            # IGNORE
            pass
        super().__init__(message)


def check_extension(filename: str, expected_extension: str) -> bool:
    """Check if the filename ends with the expected extension"""
    if not filename:
        return False
    ext_len = len(expected_extension)
    return filename[-ext_len:] == expected_extension



from copy import deepcopy
from math import floor
from typing import Dict

from .labels import clean_up as clean_up_labels


def clean_up(data: Dict) -> Dict :
    cleaned = deepcopy(data)

    annotations = cleaned.get('annotations', [])

    for annotation in annotations:
        del annotation["uid"]
        annotation["labels"] = clean_up_labels(annotation["labels"])
    
    return cleaned



# SPDX-FileCopyrightText: 2024-present Benjamin Piogé <benjamin@isahit.com>
#
# SPDX-License-Identifier: MIT

from .lab_iat_schema import LAB_IAT_SCHEMA
from .lab_iat_rectangle_schema import LAB_IAT_RECTANGLE_SCHEMA
from .lab_iat_segmentation_schema import LAB_IAT_SEGMENTATION_SCHEMA
from .lab_iat_graph_schema import LAB_IAT_GRAPH_SCHEMA
from .kili_schema import KILI_SCHEMA


__all__ = ['LAB_IAT_SCHEMA', 'LAB_IAT_RECTANGLE_SCHEMA', 'LAB_IAT_SEGMENTATION_SCHEMA', 'KILI_SCHEMA', 'LAB_IAT_GRAPH_SCHEMA']


import io
from typing import Dict, List, Optional, Union, Tuple, Generator, Iterable

from isahitlab.actions.base import BaseAction
from isahitlab.domain.dataset import (IntegrationId, FilePayload, DatasetPayload, DatasetBrowsingFilters)
from isahitlab.domain.project import ProjectId
from isahitlab.operations.batch.get_batch import GetBatchOperation
from isahitlab.operations.dataset.append_to_dataset import \
  AppendToDatasetOperation
from isahitlab.operations.dataset.create_dataset import CreateDatasetOperation
from isahitlab.operations.dataset.browse_dataset import BrowseDatasetOperation
from typeguard import typechecked


class DatasetActions(BaseAction):
    """Dataset actions"""

    @typechecked
    def create_dataset(
        self,
        project_id: ProjectId,
        name: str,
        integration_id: Optional[IntegrationId] = None,
        base_folder: Optional[str] = None,
        disable_progress_bar: Optional[bool] = False
    ) -> Dict:
        """ Create an integration

        Args:
            project_id: Project ID of the dataset
            name: Name of the dataset
            integration_id: Integration ID of the dataset
            base_folder: If integration, specify a subfolder in the integration bucket
            disable_progress_bar: Disable the progress bar display

        """

        dataset = DatasetPayload(
            name=name,
            integration_id=integration_id,
            base_folder=base_folder
        )

        return CreateDatasetOperation(self.http_client).run(project_id=project_id, dataset=dataset, disable_progress_bar=disable_progress_bar)

    @typechecked
    def browse_dataset(
        self,
        dataset_id: str,
        folder: Optional[str] = None,
        disable_progress_bar: Optional[bool] = False,
        iterate: Optional[bool] = False
    ) -> Union[Generator[Iterable[str],None,None],Iterable[str]] :
        """ Browse dataset
        
        Args:
            dataset_id: ID of the dataset
            folder: Base folder to start browsing
            disable_progress_bar: Disable the progress bar display,
            iterate: Return a generator

        Returns:
            List (or Generator) of project representations
        """

        filters = DatasetBrowsingFilters(
            dataset_id=dataset_id, 
            folder=folder if folder else None
        )
        
        operation_gen = BrowseDatasetOperation(self.http_client).run(filters=filters, disable_progress_bar=disable_progress_bar)

        if iterate:
            return operation_gen
        else:
            return list(operation_gen)
        

    @typechecked
    def append_to_dataset(
        self,
        project_id: str,
        files: List[Union[Dict, str, Tuple[str, io.IOBase]]],
        batch_id: Optional[str] = None,
        dataset_id: Optional[str] = None,
        ignore_exist_errors: Optional[bool] = False,
        disable_progress_bar: Optional[bool] = False,
    ) -> None:
        """ Add files to a dataset
        
        
        !!! info
            You must provide a `dataset_id` or a `batch_id` to use the dataset linked to the batch

        Args:
            project_id: ID of the project
            dataset_id: ID of the dataset_id
            batch_id: ID of the batch
            files: list of the file to create, str or Dict -> FilePayload(path : str, file : str)
            disable_progress_bar: Disable the progress bar display
        
        Returns:
            None
        """

        if not dataset_id and not batch_id:
            raise ValueError(
                'You must provide a dataset_id or a batch_id'
            )
        
        # Get the dataset linked to the batch
        if not dataset_id:
            batch = GetBatchOperation(self.http_client).run(batch_id=batch_id, disable_progress_bar=disable_progress_bar)
            if not batch:
                raise ValueError("Batch not found")
            if batch['projectId'] != project_id:
                raise ValueError("batch_id is not a batch of the project")
            if not batch['dataset']:
                raise ValueError('The batch has no dataset')
            dataset_id = batch['dataset']['id']    
        
        if not dataset_id:
            raise ValueError("Unknown dataset")
        
        operation = AppendToDatasetOperation(self.http_client)

        file_payloads = [*map(lambda t: FilePayload(
            file=t['file'] if not isinstance(t, str) and not isinstance(t, tuple) else t,
            path=self._format_path(t['path']) if not isinstance(t, str) and not isinstance(t, tuple) else None
            ), files)]
        
        operation.run(project_id=project_id,
                                dataset_id=dataset_id,
                                files=file_payloads,
                                ignore_exist_errors=ignore_exist_errors,
                                disable_progress_bar=disable_progress_bar
                                )
    
    def _format_path(self, path: str) -> str:
        if not path:
            return None
        path = path.strip().strip('/')
        
        if len(path) == 0:
            return None
        
        return path + '/'
        

    


from typing import List, Any, Generator

def list_duplicates(input_list : List[Any]) -> List[Any]:
    seen = set()
    dupes = []

    for x in input_list:
        if x in seen:
            dupes.append(x)
        else:
            seen.add(x)
    
    return dupes

def remove_duplicates(input_list: List[Any]) -> List[Any]:
    return list(dict.fromkeys(input_list))

def divide_chunks(input_list: List[Any], size: int) -> List[List[Any]]:
    return list(_yield_chunks(input_list, size))

def _yield_chunks(input_list: List[Any], size: int) -> Generator[List[List[Any]], None, None]:
    # looping till length l
    for i in range(0, len(input_list), size): 
        yield input_list[i:i + size]



import io
import logging
from typing import Dict, List, Optional

from isahitlab.api.dataset.api import DatasetApi
from isahitlab.domain.batch import BatchId
from isahitlab.domain.dataset import DatasetId, FilePayload
from isahitlab.domain.project import ProjectId
from isahitlab.operations.base import BaseAction
from isahitlab.exceptions import ApiBadRequest
from tqdm import tqdm
from typeguard import typechecked

logger = logging.getLogger('isahitlab.append_to_dataset')


class AppendToDatasetOperation(BaseAction):
    """Upload files to a dataset

    """


    _dataset_api : DatasetApi

    @typechecked
    def run(
        self,
        project_id: ProjectId,
        dataset_id: BatchId,
        files: List[FilePayload],
        ignore_exist_errors: Optional[bool] = False,
        disable_progress_bar: Optional[bool] = False,
    ) -> None:
        """ Add resources to a dataset
        
            project_id: ID of the project
            dataset_id: ID of the dataset_id
                You can also provide a batch_id to use the dataset linked to the batch
            batch_id: ID of the batch
            files : list of the file to create, str or Dict -> FilePayload(path : str, file : str | io.IOBase)
            disable_progress_bar: Disable the progress bar display
        """
        self._dataset_api = DatasetApi(self._http_client)
        nb_files = len(files)
        
        with tqdm(total=nb_files,  disable=disable_progress_bar, desc="Uploading files... ") as loader:
            for file in files:
                try:
                    self._upload_file(file.file, file.path, dataset_id, project_id)
                except ApiBadRequest as err:
                    if ignore_exist_errors and err.errorCode == 4002:
                        pass
                    else:
                        raise
                loader.update(1)
            
    
    def _upload_file(self, resource: str, folder: str, dataset_id: DatasetId, project_id: ProjectId) -> Dict:
        if isinstance(resource, tuple):
            file_info = self._dataset_api.upload_file(dataset_id=dataset_id, project_id=project_id, file=resource, folder=folder)
        else:
            with open(resource, 'rb') as f:
                file_info = self._dataset_api.upload_file(dataset_id=dataset_id, project_id=project_id, file=f, folder=folder)
        return file_info




from tqdm import tqdm
from typing import Optional, Dict
from isahitlab.operations.base import BaseAction
from isahitlab.domain.batch import BatchId
from isahitlab.domain.dataset import DatasetId
from isahitlab.api.batch.api import BatchApi

from typeguard import typechecked

class DeleteBatchOperation(BaseAction):
    """Delete a batch operation"""

    @typechecked
    def run(
        self,
        batch_id: BatchId,
        disable_progress_bar: Optional[bool] = False
    ) -> Dict:
        """ Delete a batch and all the tasks associated to it
        
        Args:
            batch_id: ID of the batch
            disable_progress_bar: Disable the progress bar display
        """

        batch_api = BatchApi(self._http_client)
        with tqdm(total=1,  disable=disable_progress_bar, desc="Deleting batch... ") as loader:
            updated_batch = batch_api.delete_batch(batch_id=batch_id)
            loader.update(1)

        return updated_batch
    


from typing import Dict, Optional, Union

from isahitlab.actions.base import BaseAction
from isahitlab.domain.batch import (BatchPayload, BatchId)
from isahitlab.domain.dataset import (DatasetId)
from isahitlab.operations.batch.create_batch import CreateBatchOperation
from isahitlab.operations.batch.link_dataset_to_batch import LinkDatasetToBatchOperation
from isahitlab.operations.batch.delete_batch import DeleteBatchOperation
from typeguard import typechecked


class BatchActions(BaseAction):
    """Batches actions"""

    @typechecked
    def create_batch(
        self,
        project_id: str,
        name: str,
        disable_progress_bar: Optional[bool] = False
    ) -> Dict:
        """ Create a batch in a project

        Args:
            project_id: ID of the project
            name: Name of the batch
            disable_progress_bar: Disable the progress bar display

        """

        batch = BatchPayload(name=name)

        return CreateBatchOperation(self.http_client).run(project_id=project_id, batch=batch, disable_progress_bar=disable_progress_bar)

    @typechecked
    def link_dataset_to_batch(
        self,
        batch_id: BatchId,
        dataset_id : Union[DatasetId, None],
        disable_progress_bar: Optional[bool] = False
    ) -> Dict:
        """ Link (or unlink) a dataset to a batch

        Args:
            batch_id: ID of the batch
            dataset_id : ID of the dataset to link or None to unlink a dataset from the batch
            disable_progress_bar: Disable the progress bar display

        """

        return LinkDatasetToBatchOperation(self.http_client).run(batch_id=batch_id, dataset_id=dataset_id, disable_progress_bar=disable_progress_bar)

    @typechecked
    def delete_batch(
        self,
        batch_id: BatchId,
        disable_progress_bar: Optional[bool] = False
    ) -> Dict:
        """ Delete a batch and all the tasks associated to it

        Args:
            batch_id: ID of the batch
            disable_progress_bar: Disable the progress bar display

        """

        return DeleteBatchOperation(self.http_client).run(batch_id=batch_id, disable_progress_bar=disable_progress_bar)



from tqdm import tqdm
from typing import Optional, Generator
from isahitlab.operations.base import BaseAction
from isahitlab.domain.dataset import DatasetBrowsingFilters
from isahitlab.api.dataset.api import DatasetApi

from typeguard import typechecked

class BrowseDatasetOperation(BaseAction):
    """Browse dataset operation"""

    @typechecked
    def run(
        self,
        filters : DatasetBrowsingFilters,
        disable_progress_bar: Optional[bool] = False
    ) -> Generator[tuple[str, int, int], None, None]:
        """ Browse a dataset
        
        Args:
            dataset_id: ID of the dataset
            folder : Base folder to start browsing
            disable_progress_bar: Disable the progress bar display
        """

        dataset_api = DatasetApi(self._http_client)
        with tqdm(total=0,  disable=disable_progress_bar, desc="Browsing dataset... ") as loader:            
            for (docs, loaded) in dataset_api.browse_dataset(filters=filters):
                loader.total += loaded                
                yield from docs
                loader.update(loaded)
    


import time
from typing import Dict, Generator, Union

from isahitlab.api.base import BaseApi
from isahitlab.domain.batch import BatchFilters, BatchId, BatchPayload
from isahitlab.domain.dataset import DatasetId
from isahitlab.domain.project import ProjectId

from ..helpers import Paginated, get_response_json, log_raise_for_status
from .mappers import batch_filters_mapper


class BatchApi(BaseApi):
    """Batch API Calls"""

    def get_batch_by_id(self, batch_id: BatchId) -> Dict:
        """Get batch"""

        project_configuration = self._http_client.get(
            'api/task-manager/batches/{}'.format(batch_id))

        log_raise_for_status(project_configuration)

        return get_response_json(project_configuration)

    def get_all_batches(self, filters: BatchFilters) -> Generator[tuple[Dict, int, int], None, None]:
        """Get batches"""

        page = 1
        limit = 10

        filters.pagination = True
        filters.limit = limit

        while True:
            filters.page = page

            paginated: Paginated = self._get_batches_page(filters)
            docs = paginated['docs']
            page = paginated['nextPage']

            # Count already loaded
            totalDocs = paginated['totalDocs']
            loaded = (paginated['page'] - 1) * paginated['limit'] + len(docs)

            yield (docs, loaded, totalDocs)

            if not paginated['hasNextPage']:
                break

            time.sleep(0.5)

    def _get_batches_page(self, filters: BatchFilters) -> Paginated:
        """Get batches page"""

        tasks = self._http_client.get(
            'api/task-manager/batches', params=batch_filters_mapper(filters))

        log_raise_for_status(tasks)

        return get_response_json(tasks)

    def get_batches_by_project_id(self, project_id: ProjectId) -> Dict:
        """Get batches of project"""

        project_configuration = self._http_client.get('api/task-manager/batches', params={
            "projectId": project_id
        })

        log_raise_for_status(project_configuration)

        return get_response_json(project_configuration)


    def create_batch(self, project_id: ProjectId, batch: BatchPayload):
        """Create batch in a project"""

        data = {
            "name": batch.name,
            "projectId": project_id
        }

        result = self._http_client.post('api/task-manager/batches', json=data)

        log_raise_for_status(result)

        return get_response_json(result)
    

    def link_dataset_to_batch(self, batch_id: BatchId, dataset_id : Union[DatasetId,None]):
        """Link (or unlink) a dataset to a batch"""

        data = {
            "batchId": batch_id,
            "datasetId": dataset_id
        }

        result = self._http_client.patch('api/task-manager/batches/{}/dataset'.format(batch_id), json=data)

        log_raise_for_status(result)

        return get_response_json(result)
    

    def delete_batch(self, batch_id: BatchId):
        """Delete a batch and all the tasks associated to it"""

        result = self._http_client.delete('api/task-manager/batches/{}'.format(batch_id))

        log_raise_for_status(result)

        return get_response_json(result)



from tqdm import tqdm
from typing import Optional, Dict
from isahitlab.operations.base import BaseAction
from isahitlab.domain.integration import IntegrationPayload
from isahitlab.api.integration.api import IntegrationApi

from typeguard import typechecked

class CreateIntegrationOperation(BaseAction):
    """Create integration operation"""

    @typechecked
    def run(
        self,
        integration : IntegrationPayload,
        disable_progress_bar: Optional[bool] = False
    ) -> Dict:
        """ Create an integration
        
        Args:
            project_id: ID of the batch
            integration : See IntegrationPayload
            disable_progress_bar: Disable the progress bar display
        """

        integration_api = IntegrationApi(self._http_client)
        with tqdm(total=1,  disable=disable_progress_bar, desc="Creating integration... ") as loader:
            created_integration = integration_api.create_integration(integration=integration)
            loader.update(1)

        return created_integration
    


from dataclasses import dataclass
from typing import NewType, Optional

WorkflowId = NewType("WorkflowId", str)

@dataclass
class WebhookPayload:
    """Webhook payload
    
    Args:
        url: URL to call
        header: Authorization header
    """

    url: str
    header: Optional[str] = None



from .lab_iat_schema import LAB_IAT_SCHEMA
import copy

LAB_IAT_GRAPH_SCHEMA = copy.deepcopy(LAB_IAT_SCHEMA)

# Enforce number of vertices
LAB_IAT_GRAPH_SCHEMA["properties"]["annotations"]["items"]["properties"]["polygons"]["items"]["properties"]["geometry"]["properties"]["vertices"] =  { "$ref" : "iat#/$defs/graph_vertices" }

# Enforce geometry type
LAB_IAT_GRAPH_SCHEMA["properties"]["annotations"]["items"]["properties"]["polygons"]["items"]["properties"]["geometry"]["properties"]["type"] = { "const" : "graph" }
LAB_IAT_GRAPH_SCHEMA["properties"]["annotations"]["items"]["properties"]["polygons"]["items"]["properties"]["geometry"]["properties"]["edges"] = { 
    "type": "array",
    "items": {
        "type" : "array",
        "prefixItems": [
            { "type" : "number" },
            { "type" : "number" }
        ]
    }
}
LAB_IAT_GRAPH_SCHEMA["properties"]["annotations"]["items"]["properties"]["polygons"]["items"]["properties"]["geometry"]["required"] = ["vertices", "type", "edges"]


import logging


def setup_custom_logger(name):
    logging.basicConfig()
    formatter = logging.Formatter(fmt='%(asctime)s - %(levelname)s - %(module)s - %(message)s')

    handler = logging.StreamHandler()
    handler.setFormatter(formatter)

    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)
    logger.addHandler(handler)
    return logger

